import { CosmosClient, OnomyClient } from '@onomy/client';
import { DetailedFungibleAssetBalance, OnomyHub, OnomyHubUtils } from '@onomy/hub';
import { useOnomyHub } from '@onomy/react-hub';
import { useAsyncPoll } from '@onomy/react-utils';
import { useCallback } from 'react';

async function includeVestingDetails({
  client,
  hub,
  vesting,
}: {
  client: CosmosClient;
  hub: ReturnType<typeof useOnomyHub>['hub'];
  vesting: Awaited<ReturnType<OnomyClient['getVesting']>>;
}) {
  if (!hub) throw new Error('Hub is null');
  if (!vesting) return null;

  const now = new Date().getTime() / 1000;
  const startTime = vesting.startTime.toNumber();

  const chain = client.getBlockchain();

  function coinDetails(coin: { denom: string; amount: string }) {
    return hub!.balances.includeDetails({
      amount: BigInt(coin.amount),
      asset: {
        id: coin.denom,
        chain,
      },
    });
  }

  async function getBaseVestingAccount() {
    const [delegatedFree, delegatedVesting, originalVesting] = await Promise.all([
      Promise.all(vesting!.baseVestingAccount.delegatedFree.map(coinDetails)),
      Promise.all(vesting!.baseVestingAccount.delegatedVesting.map(coinDetails)),
      Promise.all(vesting!.baseVestingAccount.originalVesting.map(coinDetails)),
    ]);
    return {
      ...vesting!,
      delegatedFree,
      delegatedVesting,
      originalVesting,
    };
  }

  async function getVestingPeriods() {
    let totalWait = 0;

    const periods =
      vesting?.vestingPeriods.map(period => {
        totalWait += period.length.toNumber();
        const unlocksAt = startTime + totalWait;
        const isUnlocked = now >= unlocksAt;
        return {
          ...period,
          totalWait,
          unlocksAt,
          isUnlocked,
        };
      }) ?? [];

    return Promise.all(
      periods.map(async period => {
        const amounts = await Promise.all(period.amount.map(coinDetails));
        return {
          ...period,
          amounts,
        };
      }) ?? []
    );
  }

  const [baseVestingAccount, vestingPeriods] = await Promise.all([
    getBaseVestingAccount(),
    getVestingPeriods(),
  ]);

  const allLockedRaw = vestingPeriods.reduce((res, period) => {
    if (!period.isUnlocked) return res;
    return [...res, ...period.amounts];
  }, [] as DetailedFungibleAssetBalance[]);

  const allLockedCombined = OnomyHub.Utils.combineBalances(allLockedRaw);
  const allLocked = await Promise.all(
    allLockedCombined.map(amount => hub.balances.includeDetails(amount))
  );

  const allLockedFiatValue = allLocked.reduce((res, item) => res + item.fiatValue, 0);
  return {
    ...vesting,
    chain,
    baseVestingAccount,
    vestingPeriods,
    allLocked,
    allLockedFiatValue,
  };
}

export function useVesting() {
  const { hub } = useOnomyHub();
  const cb = useCallback(async () => {
    if (!hub) return null;
    const clients: CosmosClient[] = [hub.chains.onomy.client];
    if (hub.chains.onex) clients.push(hub.chains.onex.client);

    const vestingDataRaw = await Promise.all(
      clients.map(async client => {
        return includeVestingDetails({
          client,
          hub,
          vesting: await client.getVesting(),
        });
      })
    );

    const vestingData = vestingDataRaw.filter((x): x is Exclude<typeof x, null> => x !== null);

    if (!vestingData.length) return null;
    const allLockedRaw = vestingData.reduce((res, item) => {
      return [...res, ...(item?.allLocked ?? [])];
    }, [] as DetailedFungibleAssetBalance[]);
    const allLockedCombined = OnomyHubUtils.combineBalances(allLockedRaw);
    const allLocked = await Promise.all(
      allLockedCombined.map(amount => hub.balances.includeDetails(amount))
    );

    const allLockedFiatValue = allLocked.reduce((res, item) => res + item.fiatValue, 0);

    return {
      allLocked,
      allLockedFiatValue,
      vestingData,
    };
  }, [hub]);

  return useAsyncPoll(cb, null);
}
